/* tsdm - Graphical login prompt
   Copyright (C) 2020, 2021 William Haddon

   This program is free software; you can redistribute it and/or modify it
   under the terms of version 2 of the GNU General Public License, as published
   by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <crypt.h>

#include "ui.h"
#include "input.h"
#include "libuser.h"

int update_text(struct input_window *w, char *data, int obfuscate) {
	size_t i, len;
	len = strlen(data);
	if (w->text_allocated) {
		free(w->text);
		w->text_allocated = 0;
	}
	if (obfuscate) {
		w->text = malloc(len + 1);
		if (!w->text)
			return OOM;
		w->text_allocated = 1;
		for (i = 0; i < len; i++)
			(w->text)[i] = '*';
		(w->text)[len] = 0;
	} else {
		w->text = data;
	}
	return SUCCESS;
}

int input_window_redraw(struct input_window *w, char *label, size_t size,
		char *data, char *message, int obfuscate) {
	xcb_rectangle_t rect, selectrect, buttonrect;
	xcb_rectangle_t labelrects[2];
	xcb_point_t pts[2];
	uint32_t values[2];
	size_t i;
	int r, lcaret, rcaret, lcaretx, rcaretx, c1x;
	r = update_text(w, data, obfuscate);
	if (r)
		return r;
	labelrects[0].x = 16;
	labelrects[0].y = w->tby;
	labelrects[0].width = w->tbx - 16;
	labelrects[0].height = 21;
	labelrects[1].x = 16;
	labelrects[1].y = 16;
	labelrects[1].width = 288;
	labelrects[1].height = w->tby - 16;
	xcb_poly_fill_rectangle(w->w.c->c, w->w.w, w->w.gow, 2, labelrects);
	if (message) {
		r = render_multiline(&(w->w), message, 16, 16, 288, w->w.bog);
		if (r)
			return r;
	}
	xcb_image_text_8(w->w.c->c, strlen(label), w->w.w, w->w.bog,
		16, w->tby + 14, label);
	rect.x = w->tbx;
	rect.y = w->tby;
	rect.width = 304 - w->tbx;
	rect.height = 20;
	xcb_set_clip_rectangles(w->w.c->c, XCB_CLIP_ORDERING_UNSORTED, w->w.bow,
		0, 0, 1, &rect);
	xcb_set_clip_rectangles(w->w.c->c, XCB_CLIP_ORDERING_UNSORTED, w->blow,
		0, 0, 1, &rect);
	xcb_set_clip_rectangles(w->w.c->c, XCB_CLIP_ORDERING_UNSORTED, w->wobl,
		0, 0, 1, &rect);
	xcb_set_clip_rectangles(w->w.c->c, XCB_CLIP_ORDERING_UNSORTED, w->w.bow,
		0, 0, 1, &rect);
	xcb_poly_fill_rectangle(w->w.c->c, w->w.w, w->w.wog, 1, &rect);
	if (w->w.flags & TEXTBOX_FOCUSED) {
		if (w->caret1 < w->caret0) {
			lcaret = w->caret1;
			rcaret = w->caret0;
		} else {
			lcaret = w->caret0;
			rcaret = w->caret1;
		}
		r = measure_text(&(w->w), w->text, lcaret, &lcaretx);
		if (r)
			goto error;
		r = measure_text(&(w->w), w->text, rcaret, &rcaretx);
		if (r)
			goto error;
		if (w->caret1 == lcaret)
			c1x = lcaretx;
		else
			c1x = rcaretx;
		if (c1x + w->tbx > 304 + w->textboxscroll)
			w->textboxscroll = c1x + w->tbx - 304;
		else if (c1x < 2 + w->textboxscroll)
			w->textboxscroll = c1x - 2;
		if (w->textboxscroll < 0)
			w->textboxscroll = 0;
		if (lcaret == rcaret) {
			pts[0].x = w->tbx + lcaretx + 2 - w->textboxscroll;
			pts[0].y = w->tby + 2;
			pts[1].x = w->tbx + lcaretx + 2 - w->textboxscroll;
			pts[1].y = w->tby + 18;
			r = render_linear_text(&(w->w), w->text, strlen(w->text),
				w->tbx + 2 - w->textboxscroll, w->tby + 14, w->w.bow);
			if (r)
				goto error;
			xcb_poly_line(w->w.c->c, XCB_COORD_MODE_ORIGIN, w->w.w, w->w.bow,
				2, pts);
		} else {
			selectrect.x = w->tbx + lcaretx + 2 - w->textboxscroll;
			selectrect.y = w->tby + 2;
			selectrect.width = rcaretx - lcaretx;
			selectrect.height = 16;
			xcb_poly_fill_rectangle(w->w.c->c, w->w.w, w->blow,
				1, &selectrect);
			r = render_linear_text(&(w->w), w->text, lcaret,
				w->tbx + 2 - w->textboxscroll, w->tby + 14, w->w.bow);
			if (r)
				goto error;
			r = render_linear_text(&(w->w), w->text + lcaret, rcaret - lcaret,
				w->tbx + lcaretx + 2 - w->textboxscroll, w->tby + 14, w->wobl);
			if (r)
				goto error;
			r = render_linear_text(&(w->w), w->text + rcaret,
				strlen(w->text) - rcaret,
				w->tbx + rcaretx + 2 - w->textboxscroll, w->tby + 14,
				w->w.bow);
			if (r)
				goto error;
		}
	} else {
		r = render_linear_text(&(w->w), w->text, strlen(w->text),
			w->tbx + 2 - w->textboxscroll, w->tby + 14, w->w.bow);
		if (r)
			goto error;
	}
	values[0] = XCB_NONE;
	xcb_change_gc(w->w.c->c, w->w.bow, XCB_GC_CLIP_MASK, values);
	xcb_change_gc(w->w.c->c, w->blow, XCB_GC_CLIP_MASK, values);
	xcb_change_gc(w->w.c->c, w->wobl, XCB_GC_CLIP_MASK, values);
	xcb_change_gc(w->w.c->c, w->w.bow, XCB_GC_CLIP_MASK, values);
	xcb_poly_rectangle(w->w.c->c, w->w.w, w->w.bog, 1, &rect);
	buttonrect.x = 128;
	buttonrect.y = w->tby + 32;
	buttonrect.width = 64;
	buttonrect.height = 28;
	draw_button(&(w->w), &buttonrect);
	if (!xcb_flush(w->w.c->c))
		return CONNECTION_LOST;
	return SUCCESS;
error:
	if (obfuscate)
		free(w->text);
	return r;
}

int init_input_window(struct connection *c, struct input_window *w) {
	int r, blue;
	r = create_window(c, &(w->w));
	if (r)
		goto error;
	w->caret0 = 0;
	w->caret1 = 0;
	w->textboxscroll = 0;
	w->tby = 16;
	w->text_allocated = 0;
	w->w.flags = TEXTBOX_FOCUSED;
	blue = create_color(c, 0, 0, 0xaa, c->screen->black_pixel);
	r = create_gc(&(w->w), blue, c->screen->white_pixel, &(w->blow));
	if (r)
		goto error;
	r = create_gc(&(w->w), c->screen->white_pixel, blue, &(w->wobl));
	if (r)
		goto error;
	if (!xcb_flush(c->c)) {
		r = CONNECTION_LOST;
		goto error;
	}
	return SUCCESS;
error:
	destroy_window(&(w->w));
	return r;
}

int get_input(struct input_window *w, char *label, size_t *size, char **data,
		char *message, int obfuscate) {
	xcb_generic_event_t *e;
	xcb_motion_notify_event_t *mne;
	xcb_button_press_event_t *bpe;
	xcb_key_press_event_t *kpe;
	xcb_client_message_event_t *cme;
	xcb_keysym_t keysym;
	char *tmp;
	size_t lcaret, rcaret, len, i;
	int r, height, lines;
	w->caret0 = 0;
	w->caret1 = 0;
	if (message) {
		r = measure_multiline(&(w->w), message, 288, &lines);
		if (r)
			return r;
		w->tby = 32 + lines * 20;
	} else {
		w->tby = 16;
	}
	r = measure_text(&(w->w), label, strlen(label), &(w->tbx));
	if (r)
		return r;
	w->tbx += 32;
	height = w->tby + 72;
	set_normal_hint_size(&(w->w), 320, height);
	xcb_configure_window(w->w.c->c, w->w.w, XCB_CONFIG_WINDOW_HEIGHT, &height);
	r = input_window_redraw(w, label, *size, *data, message, obfuscate);
	if (r)
		return r;
	for (e = xcb_wait_for_event(w->w.c->c); e;
			e = xcb_wait_for_event(w->w.c->c)) {
		if (w->caret1 < w->caret0) {
			lcaret = w->caret1;
			rcaret = w->caret0;
		} else {
			lcaret = w->caret0;
			rcaret = w->caret1;
		}
		switch (e->response_type & ~0x80) {
		case XCB_EXPOSE:
			r = input_window_redraw(w, label, *size, *data, message,
				obfuscate);
			if (r)
				goto event_error;
			break;
		case XCB_KEY_PRESS:
			kpe = (xcb_key_press_event_t *)e;
			keysym = get_keysym(&(w->w), kpe);
			len = strlen(*data);
			if (!(kpe->state & 0xc)) {
				if (keysym >= 0x20 && keysym <= 0x7e) {
					if (w->w.flags & TEXTBOX_FOCUSED) {
						if (len + lcaret + 1 - rcaret >= *size) {
							while (len + lcaret + 1 - rcaret >= *size)
								*size *= 2;
							tmp = *data;
							*data = realloc(*data, *size);
							if (!*data) {
								free(tmp);
								r = OOM;
								goto event_error;
							}
						}
						memmove(*data + lcaret + 1, *data + rcaret,
							len - rcaret);
						if (keysym >= 'A' && keysym <= 'Z' && (kpe->state & 2))
							keysym |= 0x20;
						else if (keysym >= 'a' && keysym <= 'z' &&
								(kpe->state & 2))
							keysym &= ~0x20;
						(*data)[lcaret] = keysym;
						(*data)[len + lcaret + 1 - rcaret] = 0;
						lcaret++;
						w->caret0 = lcaret;
						w->caret1 = lcaret;
						r = input_window_redraw(w, label, *size, *data,
							message, obfuscate);
						if (r)
							goto event_error;
					} else if ((w->w.flags & BUTTON_FOCUSED) &&
							keysym == 0x20) {
						w->w.flags &= ~BUTTON_FOCUSED;
						free(e);
						return SUCCESS;
					}
				} else {
					switch (keysym) {
					case 0xfe20: /* Shift+Tab */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							w->w.flags &= ~TEXTBOX_FOCUSED;
							w->w.flags &= ~BUTTON_FOCUSED;
						} else if (w->w.flags & BUTTON_FOCUSED) {
							w->w.flags &= ~BUTTON_FOCUSED;
							w->w.flags |= TEXTBOX_FOCUSED;
							w->caret0 = 0;
							w->caret1 = 0;
						} else {
							w->w.flags |= BUTTON_FOCUSED;
						}
						r = input_window_redraw(w, label, *size, *data,
							message, obfuscate);
						if (r)
							goto event_error;
						break;
					case 0xff08: /* Backspace */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							if (lcaret == rcaret) {
								if (lcaret) {
									memmove(*data + lcaret - 1, *data + lcaret,
										len - lcaret);
									(*data)[len-1] = 0;
									lcaret--;
								}
							} else {
								memmove(*data + lcaret, *data + rcaret,
									len - rcaret);
								(*data)[len + lcaret - rcaret] = 0;
							}
							w->caret0 = lcaret;
							w->caret1 = lcaret;
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					case 0xff09: /* Tab */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							w->w.flags &= ~TEXTBOX_FOCUSED;
							w->w.flags |= BUTTON_FOCUSED;
						} else if (w->w.flags & BUTTON_FOCUSED) {
							w->w.flags &= ~BUTTON_FOCUSED;
						} else {
							w->w.flags |= TEXTBOX_FOCUSED;
							w->caret0 = 0;
							w->caret1 = 0;
						}
						r = input_window_redraw(w, label, *size, *data,
							message, obfuscate);
						if (r)
							goto event_error;
						break;
					case 0xff0d: /* Return */
						w->w.flags &= ~BUTTON_FOCUSED;
						free(e);
						return SUCCESS;
					case 0xff50: /* Home */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							w->caret1 = 0;
							if (!(kpe->state & 1))
								w->caret0 = w->caret1;
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					case 0xff51: /* Left */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							if (kpe->state & 1) {
								if (w->caret1)
									w->caret1--;
							} else {
								if (lcaret && lcaret == rcaret)
									lcaret--;
								w->caret0 = lcaret;
								w->caret1 = lcaret;
							}
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					case 0xff53: /* Right */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							if (kpe->state & 1) {
								if (w->caret1 < len)
									w->caret1++;
							} else {
								if (rcaret < len && lcaret == rcaret)
									rcaret++;
								w->caret0 = rcaret;
								w->caret1 = rcaret;
							}
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					case 0xff57: /* End */
						if (w->w.flags & TEXTBOX_FOCUSED) {
							w->caret1 = len;
							if (!(kpe->state & 1))
								w->caret0 = w->caret1;
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					case 0xffff: /* Delete */
						if ((w->w.flags & TEXTBOX_FOCUSED) && len) {
							if (lcaret == rcaret) {
								if (lcaret < len) {
									memmove(*data + lcaret, *data + lcaret + 1,
										len - lcaret - 1);
									(*data)[len-1] = 0;
								}
							} else {
								memmove(*data + lcaret, *data + rcaret,
									len - rcaret);
								(*data)[len + lcaret - rcaret] = 0;
							}
							w->caret0 = lcaret;
							w->caret1 = lcaret;
							r = input_window_redraw(w, label, *size, *data,
								message, obfuscate);
							if (r)
								goto event_error;
						}
						break;
					}
				}
			}
			break;
		case XCB_BUTTON_PRESS:
			bpe = (xcb_button_press_event_t *)e;
			if (bpe->event_x >= 128 && bpe->event_x <= 192 &&
					bpe->event_y >= w->tby + 32 &&
					bpe->event_y <= w->tby + 60) {
				w->w.flags |= BUTTON_HELD;
				r = input_window_redraw(w, label, *size, *data, message,
					obfuscate);
				if (r)
					goto event_error;
			} else if (bpe->event_x >= 88 && bpe->event_x <= 304 &&
					bpe->event_y >= w->tby && bpe->event_y <= w->tby + 20) {
				w->w.flags |= TEXTBOX_FOCUSED | TEXTBOX_HELD;
				r = char_at_width(&(w->w), w->text,
					bpe->event_x + w->textboxscroll - w->tbx - 2,
					&(w->caret0));
				if (r)
					goto event_error;
				w->caret1 = w->caret0;
				r = input_window_redraw(w, label, *size, *data, message,
					obfuscate);
				if (r)
					goto event_error;
			}
			break;
		case XCB_BUTTON_RELEASE:
			bpe = (xcb_button_press_event_t *)e;
			if ((w->w.flags & BUTTON_HELD) &&
					bpe->event_x >= 128 && bpe->event_x <= 192 &&
					bpe->event_y >= w->tby + 32 &&
					bpe->event_y <= w->tby + 60) {
				w->w.flags &= ~TEXTBOX_FOCUSED;
				free(e);
				return SUCCESS;
			} else if (w->w.flags & TEXTBOX_HELD) {
				r = char_at_width(&(w->w), w->text,
					bpe->event_x + w->textboxscroll - w->tbx - 2,
					&(w->caret1));
				if (r)
					goto event_error;
				w->w.flags &= ~BUTTON_FOCUSED;
			} else {
				w->w.flags &= ~TEXTBOX_FOCUSED;
				w->w.flags &= ~BUTTON_FOCUSED;
			}
			w->w.flags &= ~BUTTON_HELD;
			w->w.flags &= ~TEXTBOX_HELD;
			r = input_window_redraw(w, label, *size, *data, message,
				obfuscate);
			if (r)
				goto event_error;
			break;
		case XCB_MOTION_NOTIFY:
			mne = (xcb_motion_notify_event_t *)e;
			if (w->w.flags & TEXTBOX_HELD) {
				r = char_at_width(&(w->w), w->text,
					mne->event_x + w->textboxscroll - w->tbx - 2,
					&(w->caret1));
				if (r)
					goto event_error;
				r = input_window_redraw(w, label, *size, *data, message,
					obfuscate);
				if (r)
					goto event_error;
			} else if (mne->event_x >= 128 && mne->event_x <= 192 &&
					mne->event_y >= w->tby + 32 &&
					mne->event_y <= w->tby + 60) {
				if (!(w->w.flags & BUTTON_HOVER)) {
					w->w.flags |= BUTTON_HOVER;
					r = input_window_redraw(w, label, *size, *data, message,
						obfuscate);
					if (r)
						goto event_error;
				}
			} else {
				if (w->w.flags & BUTTON_HOVER) {
					w->w.flags &= ~BUTTON_HOVER;
					r = input_window_redraw(w, label, *size, *data, message,
						obfuscate);
					if (r)
						goto event_error;
				}
			}
			break;
		case XCB_LEAVE_NOTIFY:
			if (w->w.flags & BUTTON_HOVER) {
				w->w.flags &= ~BUTTON_HOVER;
				r = input_window_redraw(w, label, *size, *data, message,
					obfuscate);
				if (r)
					goto event_error;
			}
			break;
		case XCB_CLIENT_MESSAGE:
			cme = (xcb_client_message_event_t *)e;
			if (cme->format == 32 && cme->type == w->w.c->wm_protocols &&
					cme->data.data32[0] == w->w.c->wm_delete_window) {
				free(e);
				return WINDOW_CLOSED;
			}
			break;
		case 0:
			free(e);
			return GRAPHICS_ERROR;
		}
		free(e);
	}
	return CONNECTION_LOST;
event_error:
	free(e);
	return r;
}
