/* tsdm - Graphical login prompt
   Copyright (C) 2020, 2021 William Haddon

   This program is free software; you can redistribute it and/or modify it
   under the terms of version 2 of the GNU General Public License, as published
   by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; withou even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <shadow.h>
#include <crypt.h>

#include "ui.h"
#include "input.h"
#include "libuser.h"

char **environ;

pid_t vfe(char *proc, char **argv, struct connection *c) {
	pid_t r;
	int i, fd;
	errno = 0;
	/* Signal handlers should be disabled, rathen than actually blocking
	   signals, so child processes do not start with signals blocked and keep
	   signals blocked. It is not safe for the child process to act on signals
	   as if it is the parent. However, signals will go to the signal FD and
	   not disrupt the execution of the child. */
	r = vfork();
	if (r) {
		if (errno)
			r = -1;
	} else {
		fd = xcb_get_file_descriptor(c->c);
		if (fd >= 0)
			close(fd);
		if (execve(proc, argv, environ))
			_exit(1);
	}
	return r;
}

int login(struct connection *c, char *username) {
	struct passwd *pw;
	struct group *gr;
	gid_t gids[NGROUPS_MAX];
	char pwd[PATH_MAX];
	char *env[8];
	char *argv[2];
	char *display, *msg, *msg2, **s;
	size_t i;
	int r, pwdvalid;
	pwdvalid = !!getcwd(pwd, PATH_MAX);
	argv[0] = "/etc/init/xprelogin";
	argv[1] = NULL;
	if (vfe("/etc/init/xprelogin", argv, c) < 0) {
		printf("Error executing /etc/init/xprelogin: %s\n", strerror(errno));
		return message_window(c, "Error logging in");
	}
	pw = getpwnam(username);
	if (!pw) {
		printf("User account not found after verifying credentials\n");
		return message_window(c, "Error logging in");
	}
	errno = 0;
	i = 0;
	for (gr = getgrent(); gr; gr = getgrent()) {
		for (s = gr->gr_mem; *s; s++) {
			if (!strcmp(*s, username)) {
				if (i >= NGROUPS_MAX) {
					printf("Warning: "
						"maximum number of supplemental groups exceeded\n");
					goto groupsdone;
				}
				gids[i] = gr->gr_gid;
				i++;
			}
		}
	}
	if (errno) {
		printf("Warning: error reading groups file: %s\n", strerror(errno));
		errno = 0;
	}
	if (i >= NGROUPS_MAX) {
		printf("Warning: maximum number of supplemental groups exceeded\n");
	} else {
		gids[i] = pw->pw_gid;
		i++;
	}
groupsdone:
	endgrent();
	env[0] = malloc(6 + strlen(pw->pw_dir));
	if (!env[0]) {
		printf("Out of memory\n");
		return message_window(c, "Error logging in");
	}
	sprintf(env[0], "HOME=%s", pw->pw_dir);
	env[1] = malloc(7 + strlen(pw->pw_shell));
	if (!env[1]) {
		printf("Out of memory\n");
		r = message_window(c, "Error logging in");
		goto error;
	}
	sprintf(env[1], "SHELL=%s", pw->pw_shell);
	env[2] = malloc(6 + strlen(pw->pw_name));
	if (!env[2]) {
		printf("Out of memory\n");
		r = message_window(c, "Error logging in");
		goto error1;
	}
	sprintf(env[2], "USER=%s", pw->pw_name);
	env[3] = malloc(9 + strlen(pw->pw_name));
	if (!env[3]) {
		printf("Out of memory\n");
		r = message_window(c, "Error logging in");
		goto error2;
	}
	sprintf(env[3], "LOGNAME=%s", pw->pw_name);
	env[4] = "PATH=/bin:/sbin:/usr/bin:/usr/sbin";
	env[5] = "IFS= \t\n";
	display = getenv("DISPLAY");
	if (display) {
		env[6] = malloc(9 + strlen(display));
		if (!env[6]) {
			printf("Out of memory\n");
			r = message_window(c, "Error logging in");
			goto error3;
		}
		sprintf(env[6], "DISPLAY=%s", display);
	} else {
		env[6] = "DISPLAY=";
	}
	env[7] = NULL;
	/* Set the supplemental GIDs */
	if (setgroups(i, gids)) {
		printf("Error switching supplemental GIDs: %s\n", strerror(errno));
		r = message_window(c, "Error logging in");
		goto error4;
	}
	/* Change user. setgid must be called before setuid because after setuid we
	   are no longer root. */
	if (setgid(pw->pw_gid)) {
		printf("Error switching GID: %s\n", strerror(errno));
		r = message_window(c, "Error logging in");
		goto error4;
	}
	if (setuid(pw->pw_uid)) {
		printf("Error switching user: %s\n", strerror(errno));
		r = message_window(c, "Error logging in");
		goto error4;
	}
	if (chdir(pw->pw_dir)) {
		msg = strerror(errno);
		printf("Warning: can not switch to home directory %s: %s\n",
			pw->pw_dir, msg);
		msg2 = malloc(strlen("Warning: can not switch to home directory : ") +
			strlen(pw->pw_dir) + strlen(msg) + 1);
		if (msg2) {
			sprintf(msg2, "Warning: can not switch to home directory %s: %s",
				pw->pw_dir, msg);
			r = message_window(c, msg2);
			free(msg2);
		}
		if (r)
			goto error4;
	}
	argv[0] = "/etc/init/xlogin";
	argv[1] = NULL;
	if (execve("/etc/init/xlogin", argv, env)) {
		if (pwdvalid)
			chdir(pwd);
		printf("Error executing /etc/init/xlogin: %s: %s\n", pw->pw_dir,
			strerror(errno));
		r = message_window(c, "Error logging in");
		goto error4;
	}
error4:
	free(env[6]);
error3:
	free(env[3]);
error2:
	free(env[2]);
error1:
	free(env[1]);
error:
	free(env[0]);
	return r;
}

char *progname;

int main(int argc, char **argv) {
	struct connection c;
	struct input_window w;
	char msg[64];
	char *username, *password, *password2, *password3;
	size_t username_size, password_size, password2_size, password3_size;
	int r, arg;
	progname = argv[0];
	r = init_connection(&c);
	if (r)
		goto error;
	r = init_input_window(&c, &w);
	if (r)
		goto error;
	username_size = 256;
	username = malloc(username_size);
	if (!username)
		goto oom;
	username[0] = 0;
	password_size = 256;
	password = malloc(password_size);
	if (!password)
		goto oom;
	password[0] = 0;
	password2_size = 256;
	password2 = malloc(password2_size);
	if (!password2)
		goto oom;
	password2[0] = 0;
	password3_size = 256;
	password3 = malloc(password3_size);
	if (!password3)
		goto oom;
	password3[0] = 0;
	r = get_input(&w, "Username:", &username_size, &username, NULL, 0);
	if (r)
		goto error;
	if (check_username(username) != SKIP_PASSWORD) {
		r = get_input(&w, "Password:", &password_size, &password, NULL, 1);
		if (r)
			goto error;
	}
	destroy_window(&(w.w));
	r = check_credentials(username, password, &arg);
	switch (r) {
	case CHANGE_PASSWORD:
		r = init_input_window(&c, &w);
		if (r)
			goto error;
		password2[0] = 0;
		r = get_input(&w, "New password:", &password2_size, &password2,
			"Your password has expired. Choose a new password.", 1);
		if (r)
			goto error;
		password3[0] = 0;
		r = get_input(&w, "New password again:", &password3_size, &password3,
			"Your password has expired. Choose a new password.", 1);
		if (r)
			goto error;
		destroy_window(&(w.w));
		/* We need to check the password again after all input is received so
			users can not avoid authenticating with the second password by
			delaying input after another process changes the password */
		if (check_username(username) != SKIP_PASSWORD) {
			r = check_credentials(username, password, NULL);
			if (r != VALID && r != CHANGE_PASSWORD && r != PASSWORD_WARNING)
				goto error;
		}
		if (strcmp(password2, password3)) {
			r = message_window(&c, "New passwords don't match.");
			goto error;
		} else if (!strcmp(password, password2)) {
			r = message_window(&c, "New password doesn't differ from old.");
			goto error;
		} else {
			r = change_password(username, password2);
			if (r)
				goto error;
		}
		break;
	case PASSWORD_WARNING:
		sprintf(msg, "Warning: password will expire in %d days.", arg);
		r = message_window(&c, msg);
	case VALID:
		break;
	default:
		goto error;
	}
	r = login(&c, username);
error:
	switch (r) {
	case INVALID_PASSWORD:
	case INVALID_USER:
		r = message_window(&c, "Invalid credentials\n");
		break;
	case LOCKED:
		r = message_window(&c, "Account is locked\n");
		break;
	case PASSWORD_EXPIRED:
		r = message_window(&c, "Password is expired\n");
		break;
	case ACCOUNT_EXPIRED:
		r = message_window(&c, "Account is expired\n");
		break;
	case NO_ENTROPY:
		fprintf(stderr, "%s: unexpected failure: no entropy\n");
		r = message_window(&c, "Unexpected failure\n");
	case ERROR:
	case FILE_MISSING:
	case FILE_BUSY:
	case PERMISSION:
		fprintf(stderr, "%s: unexpected failure: %s\n", progname,
			strerror(errno));
		r = message_window(&c, "Unexpected failure\n");
		break;
	}
	switch (r) {
	case WINDOW_CLOSED:
		return 0;
	case CONNECTION_LOST:
		fprintf(stderr, "%s: X11 connection lost\n", progname);
		return 1;
	case GRAPHICS_ERROR:
		fprintf(stderr, "%s: X11 error\n", progname);
		return 1;
	case CONNECTION_FAILED:
		fprintf(stderr, "%s: X11 connection failed\n", progname);
		return 1;
	case FONT_ERROR:
		fprintf(stderr, "%s: X11: font not found\n", progname);
		return 1;
	case OOM:
oom:
		fprintf(stderr, "%s: out of memory\n", progname);
		return 1;
	}
	return 0;
}
