/* tsdm - Graphical login prompt
   Copyright (C) 2020, 2021 William Haddon

   This program is free software; you can redistribute it and/or modify it
   under the terms of version 2 of the GNU General Public License, as published
   by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <crypt.h>
#include <pwd.h>
#include <grp.h>

#include "libuser.h"
#include "ui.h"

int create_gc(struct window *w, int fg, int bg, xcb_gcontext_t *result) {
	xcb_generic_error_t *err;
	uint32_t values[3];
	*result = xcb_generate_id(w->c->c);
	values[0] = fg;
	values[1] = bg;
	values[2] = w->c->font;
	err = xcb_request_check(w->c->c, xcb_create_gc_checked(w->c->c, *result,
		w->w, XCB_GC_FOREGROUND|XCB_GC_BACKGROUND|XCB_GC_FONT, values));
	if (err) {
		if (err->error_code == XCB_ALLOC)
			return OOM;
		else
			return GRAPHICS_ERROR;
	} else {
		return SUCCESS;
	}
}

int create_color(struct connection *c, int red, int green, int blue,
		int backup) {
	xcb_generic_error_t *err;
	xcb_alloc_color_reply_t *acr;
	int result;
	err = NULL;
	acr = xcb_alloc_color_reply(c->c, xcb_alloc_color(c->c,
		c->screen->default_colormap, red * 257, green * 257, blue * 257),
		&err);
	if (err) {
		free(err);
		result = backup;
	} else {
		result = acr->pixel;
	}
	free(acr);
	return result;
}

int measure_text(struct window *w, char *s, size_t pos, int *width) {
	xcb_generic_error_t *err;
	xcb_query_text_extents_reply_t *qtxr;
	xcb_char2b_t *copy;
	size_t i;
	int result;
	copy = malloc(pos * sizeof(xcb_char2b_t));
	if (!copy)
		return OOM;
	for (i = 0; i < pos; i++) {
		copy[i].byte1 = 0;
		copy[i].byte2 = s[i];
	}
	err = NULL;
	qtxr = xcb_query_text_extents_reply(w->c->c,
		xcb_query_text_extents(w->c->c, w->c->font, pos, copy), &err);
	if (err) {
		free(err);
		result = GRAPHICS_ERROR;
	} else {
		*width = qtxr->overall_width;
		result = SUCCESS;
	}
	free(qtxr);
	free(copy);
	return result;
}

int measure_multiline(struct window *w, char *s, int width, int *lines) {
	char *t, *prev, *begin;
	int r, x;
	*lines = 0;
	prev = s;
	begin = s;
	for (t = s; *t; t++) {
		switch (*t) {
		case ' ':
		case '\t':
			r = measure_text(w, begin, t - begin, &x);
			if (r)
				return r;
			if (x > width) {
				if (prev == begin)
					begin = t + 1;
				else
					begin = prev;
				(*lines)++;
			}
			prev = t + 1;
			break;
		case '\n':
			r = measure_text(w, begin, t - begin, &x);
			if (r)
				return r;
			if (x > width && prev != begin)
				(*lines)++;
			(*lines)++;
			begin = t + 1;
			prev = t + 1;
			break;
		}
	}
	r = measure_text(w, begin, t - begin, &x);
	if (r)
		return r;
	if (x > width && prev != begin)
		(*lines)++;
	(*lines)++;
	return SUCCESS;
}

int char_at_width(struct window *w, char *s, int width, size_t *length) {
	size_t len, unit;
	int sw, r;
	len = strlen(s);
	*length = 0;
	for (unit = 1; unit * 2 <= len; unit *= 2);
	for (; unit; unit /= 2) {
		if (*length + unit <= len) {
			r = measure_text(w, s, *length + unit, &sw);
			if (r)
				return r;
			if (sw < width)
				*length += unit;
		}
	}
	return SUCCESS;
}

xcb_atom_t get_atom(xcb_connection_t *c, char *s) {
	xcb_generic_error_t *err;
	xcb_intern_atom_reply_t *xiar;
	xcb_atom_t result;
	err = NULL;
	xiar = xcb_intern_atom_reply(c, xcb_intern_atom(c, 0, strlen(s), s), &err);
	if (err) {
		free(err);
		result = XCB_ATOM_NONE;
	} else {
		result = xiar->atom;
	}
	free(xiar);
	return result;
}

void destroy_window(struct window *w) {
	xcb_destroy_window(w->c->c, w->w);
}

/* my X server seems to reject strings longer than 256 characters */
int render_linear_text(struct window *w, char *s, size_t len, int x, int y,
		xcb_gcontext_t gc) {
	size_t i, chunk;
	int width, tw, r;
	tw = 0;
	for (i = 0; i < len; i += 255) {
		if (i + 255 < len)
			chunk = 255;
		else
			chunk = len - i;
		xcb_image_text_8(w->c->c, chunk, w->w, gc, x + tw, y, s + i);
		r = measure_text(w, s + i, chunk, &width);
		if (r)
			return r;
		tw += width;
	}
	return SUCCESS;
}

int render_multiline(struct window *w, char *s, int x, int y, int width,
		xcb_gcontext_t gc) {
	char *t, *prev, *begin;
	int r, cx;
	prev = s;
	begin = s;
	y += 14;
	for (t = s; *t; t++) {
		switch (*t) {
		case ' ':
		case '\t':
			r = measure_text(w, begin, t - begin, &cx);
			if (r)
				return r;
			if (cx > width) {
				if (prev == begin) {
					r = render_linear_text(w, begin, t - begin, x, y, gc);
					begin = t + 1;
				} else {
					r = render_linear_text(w, begin, prev - begin, x, y, gc);
					begin = prev;
				}
				y += 20;
			}
			prev = t + 1;
			break;
		case '\n':
			r = measure_text(w, begin, t - begin, &cx);
			if (r)
				return r;
			if (cx > width && prev != begin) {
				r = render_linear_text(w, begin, prev - begin, x, y, gc);
				y += 20;
				begin = prev;
			}
			r = render_linear_text(w, begin, t - begin, x, y, gc);
			y += 20;
			begin = t + 1;
			prev = t + 1;
			break;
		}
	}
	r = measure_text(w, begin, t - begin, &cx);
	if (r)
		return r;
	if (cx > width && prev != begin) {
		r = render_linear_text(w, begin, prev - begin, x, y, gc);
		begin = prev;
		y += 20;
	}
	r = render_linear_text(w, begin, t - begin, x, y, gc);
	y += 20;
	return SUCCESS;
}

void draw_button(struct window *w, xcb_rectangle_t *rect) {
	if (w->flags & (BUTTON_HOVER | BUTTON_FOCUSED))
		xcb_poly_fill_rectangle(w->c->c, w->w, w->wog, 1, rect);
	else
		xcb_poly_fill_rectangle(w->c->c, w->w, w->gow, 1, rect);
	xcb_poly_rectangle(w->c->c, w->w, w->bog, 1, rect);
	if (w->flags & (BUTTON_HOVER | BUTTON_FOCUSED))
		xcb_image_text_8(w->c->c, 2, w->w, w->bow, 156, rect->y + 19, "OK");
	else
		xcb_image_text_8(w->c->c, 2, w->w, w->bog, 156, rect->y + 19, "OK");
}

int init_connection(struct connection *c) {
	xcb_generic_error_t *err;
	xcb_screen_iterator_t roots;
	int r, i, scrno;
	c->c = xcb_connect(NULL, &scrno);
	if (xcb_connection_has_error(c->c))
		return CONNECTION_FAILED;
	c->setup = xcb_get_setup(c->c);
	err = NULL;
	c->gkmr = xcb_get_keyboard_mapping_reply(c->c,
		xcb_get_keyboard_mapping(c->c, c->setup->min_keycode,
		c->setup->max_keycode - c->setup->min_keycode), &err);
	if (err) {
		r = GRAPHICS_ERROR;
		free(err);
		goto error;
	}
	c->keysyms = xcb_get_keyboard_mapping_keysyms(c->gkmr);
	roots = xcb_setup_roots_iterator(c->setup);
	for (i = 0; i < scrno; i++)
		xcb_screen_next(&roots);
	c->screen = roots.data;
	c->wm_protocols = get_atom(c->c, "WM_PROTOCOLS");
	if (c->wm_protocols == XCB_ATOM_NONE)
		goto error;
	c->wm_delete_window = get_atom(c->c, "WM_DELETE_WINDOW");
	if (c->wm_delete_window == XCB_ATOM_NONE) {
		r = GRAPHICS_ERROR;
		goto error;
	}
	c->font = xcb_generate_id(c->c);
	err = xcb_request_check(c->c,
		xcb_open_font_checked(c->c, c->font, 4, "6x13"));
	if (err) {
		if (err->error_code == XCB_NAME)
			r = FONT_ERROR;
		else
			r = GRAPHICS_ERROR;
		free(err);
		goto error;
	}
	return SUCCESS;
error:
	free(c->gkmr);
	xcb_disconnect(c->c);
	return r;
}

void destroy_connection(struct connection *c) {
	free(c->gkmr);
	xcb_disconnect(c->c);
}

void set_normal_hint_size(struct window *w, int width, int height) {
	uint32_t normal_hints[18];
	int i;
	normal_hints[0] = 0x38;
	for (i = 1; i < 5; i++)
		normal_hints[i] = 0;
	normal_hints[5] = width;
	normal_hints[6] = height;
	normal_hints[7] = width;
	normal_hints[8] = height;
	normal_hints[9] = width;
	normal_hints[10] = height;
	for (i = 11; i < 15; i++)
		normal_hints[i] = 0;
	normal_hints[16] = width;
	normal_hints[17] = height;
	normal_hints[18] = XCB_GRAVITY_NORTH_WEST;
	xcb_change_property(w->c->c, XCB_PROP_MODE_REPLACE, w->w,
		XCB_ATOM_WM_NORMAL_HINTS, XCB_ATOM_WM_SIZE_HINTS, 32, 18,
		normal_hints);
}

int create_window(struct connection *c, struct window *w) {
	uint32_t normal_hints[18];
	uint32_t values[2];
	int r, i, gray;
	w->c = c;
	w->flags = 0;
	gray = create_color(c, 0xaa, 0xaa, 0xaa, c->screen->white_pixel);
	w->w = xcb_generate_id(c->c);
	values[0] = gray;
	values[1] = XCB_EVENT_MASK_KEY_PRESS | XCB_EVENT_MASK_BUTTON_PRESS |
		XCB_EVENT_MASK_BUTTON_RELEASE | XCB_EVENT_MASK_POINTER_MOTION |
		XCB_EVENT_MASK_EXPOSURE | XCB_EVENT_MASK_LEAVE_WINDOW;
	xcb_create_window(c->c, XCB_COPY_FROM_PARENT, w->w, c->screen->root,
		0, 0, 320, 88, 0, XCB_WINDOW_CLASS_INPUT_OUTPUT, XCB_COPY_FROM_PARENT,
		XCB_CW_BACK_PIXEL|XCB_CW_EVENT_MASK, values);
	xcb_change_property(c->c, XCB_PROP_MODE_REPLACE, w->w, XCB_ATOM_WM_NAME,
		XCB_ATOM_STRING, 8, 6, "Log in");
	xcb_change_property(c->c, XCB_PROP_MODE_REPLACE, w->w, c->wm_protocols,
		XCB_ATOM_ATOM, 32, 1, &(c->wm_delete_window));
	set_normal_hint_size(w, 320, 88);
	xcb_map_window(c->c, w->w);
	r = create_gc(w, c->screen->black_pixel, gray, &(w->bog));
	if (r)
		goto error;
	r = create_gc(w, c->screen->white_pixel, gray, &(w->wog));
	if (r)
		goto error;
	r = create_gc(w, c->screen->black_pixel, c->screen->white_pixel,
		&(w->bow));
	if (r)
		goto error;
	r = create_gc(w, gray, c->screen->white_pixel, &(w->gow));
	if (r)
		goto error;
	return SUCCESS;
error:
	destroy_window(w);
	return r;
}

xcb_keysym_t get_keysym(struct window *w, xcb_key_press_event_t *kpe) {
	xcb_keysym_t keysym;
	size_t i;
	int count;
	count = w->c->gkmr->keysyms_per_keycode;
	i = 0;
	if (kpe->state & 1)
		i |= 1;
	if (kpe->state & 0x80)
		i |= 4;
	keysym = w->c->keysyms[
		(kpe->detail - w->c->setup->min_keycode) * count + i];
	if (!keysym)
		keysym = w->c->keysyms[
			(kpe->detail - w->c->setup->min_keycode) * count];
	if (kpe->state & 0xc)
		keysym = 0;
	return keysym;
}

struct message_window {
	struct window w;
	int height;
};

int message_window_redraw(struct message_window *w, char *message) {
	xcb_rectangle_t rect;
	int r;
	r = render_multiline(&(w->w), message, 16, 16, 288, w->w.bog);
	if (r)
		return r;
	rect.x = 128;
	rect.y = w->height - 48;
	rect.width = 64;
	rect.height = 28;
	draw_button(&(w->w), &rect);
	if (!xcb_flush(w->w.c->c))
		return CONNECTION_LOST;
	return SUCCESS;
}

int message_window(struct connection *c, char *message) {
	struct message_window w;
	xcb_generic_error_t *err;
	uint32_t normal_hints[18];
	uint32_t values[2];
	xcb_generic_event_t *e;
	xcb_motion_notify_event_t *mne;
	xcb_button_press_event_t *bpe;
	xcb_key_press_event_t *kpe;
	xcb_client_message_event_t *cme;
	xcb_keysym_t keysym;
	int r, i, gray, lines;
	r = create_window(c, &(w.w));
	if (r)
		goto done;
	r = measure_multiline(&(w.w), message, 288, &lines);
	if (r)
		goto done;
	w.height = 72 + lines * 20;
	set_normal_hint_size(&(w.w), 320, w.height);
	xcb_configure_window(c->c, w.w.w, XCB_CONFIG_WINDOW_HEIGHT, &(w.height));
	if (r)
		return r;
	if (!xcb_flush(c->c)) {
		r = CONNECTION_LOST;
		goto done;
	}
	r = message_window_redraw(&w, message);
	if (r)
		goto done;
	for (e = xcb_wait_for_event(c->c); e; e = xcb_wait_for_event(c->c)) {
		switch (e->response_type & ~0x80) {
		case XCB_EXPOSE:
			r = message_window_redraw(&w, message);
			if (r)
				goto done2;
			break;
		case XCB_KEY_PRESS:
			kpe = (xcb_key_press_event_t *)e;
			keysym = get_keysym(&(w.w), kpe);
			if (!(kpe->state & 0xc)) {
				switch (keysym) {
				case 0x20: /* Space */
					if (w.w.flags & BUTTON_FOCUSED) {
						r = SUCCESS;
						goto done2;
					}
					break;
				case 0xfe20: /* Shift+Tab */
				case 0xff09: /* Tab */
					w.w.flags ^= BUTTON_FOCUSED;
					r = message_window_redraw(&w, message);
					if (r)
						goto done2;
					break;
				case 0xff0d: /* Return */
					r = SUCCESS;
					goto done2;
					break;
				}
			}
			break;
		case XCB_BUTTON_PRESS:
			bpe = (xcb_button_press_event_t *)e;
			if (bpe->event_x >= 128 && bpe->event_x <= 192 &&
					bpe->event_y + 48 >= w.height &&
					bpe->event_y + 20 <= w.height) {
				w.w.flags |= BUTTON_HELD;
				r = message_window_redraw(&w, message);
				if (r)
					goto done2;
			}
			break;
		case XCB_BUTTON_RELEASE:
			bpe = (xcb_button_press_event_t *)e;
			if ((w.w.flags & BUTTON_HELD) &&
					bpe->event_x >= 128 && bpe->event_x <= 192 &&
					bpe->event_y + 48 >= w.height &&
					bpe->event_y + 20 <= w.height) {
				r = SUCCESS;
				goto done2;
			} else {
				w.w.flags &= ~BUTTON_FOCUSED;
			}
			w.w.flags &= ~BUTTON_HELD;
			r = message_window_redraw(&w, message);
			if (r)
				goto done2;
			break;
		case XCB_MOTION_NOTIFY:
			mne = (xcb_motion_notify_event_t *)e;
			if (mne->event_x >= 128 && mne->event_x <= 192 &&
					mne->event_y + 48 >= w.height &&
					mne->event_y + 20 <= w.height) {
				if (!(w.w.flags & BUTTON_HOVER)) {
					w.w.flags |= BUTTON_HOVER;
					r = message_window_redraw(&w, message);
					if (r)
						goto done2;
				}
			} else if (w.w.flags & BUTTON_HOVER) {
				w.w.flags &= ~BUTTON_HOVER;
				r = message_window_redraw(&w, message);
				if (r)
					goto done2;
			}
			break;
		case XCB_LEAVE_NOTIFY:
			if (w.w.flags & BUTTON_HOVER) {
				w.w.flags &= ~BUTTON_HOVER;
				r = message_window_redraw(&w, message);
				if (r)
					goto done2;
			}
			break;
		case XCB_CLIENT_MESSAGE:
			cme = (xcb_client_message_event_t *)e;
			if (cme->format == 32 && cme->type == c->wm_protocols &&
					cme->data.data32[0] == c->wm_delete_window) {
				r = WINDOW_CLOSED;
				goto done2;
			}
			break;
		case 0:
			r = GRAPHICS_ERROR;
			goto done2;
		}
		free(e);
	}
	r = CONNECTION_LOST;
done2:
	free(e);
done:
	destroy_window(&(w.w));
	return r;
}
